from functools import wraps
import random
import pyglet
from pyglet.window import key

from scenegraph.objloader import ModelLoader
from scenegraph.scenegraph import ShaderGroup, v3, ModelNode
from scenegraph.shaderlib import textured_specular
from scenegraph.particles import ParticleSystemNode
from scenegraph.lighting import Light

from spells import Fire, MagicMissile, Spell, Shield, Laser, Invisibility

loader = ModelLoader()

LIGHT_POS = -100, 50, 30, 0

GRAVITY = 20

wizard_model = loader.load_animation('wizard',
    sequences={
        'stand': [0],
        'draw': range(10),
        'ready': [10],
        'cast': range(10, 20),
        'stand-down': range(10, 0, -1),
        'hit': range(20, 26, 3) + range(25, 20, -1),
        'death': range(26, 36),
        'dead': [35]
    },
    default='stand',
    next={
        'draw': 'ready',
        'cast': 'ready',
        'stand-down': 'stand',
        'hit': 'cast',
        'death': 'dead'
    }
)
wizard2_model = wizard_model.copy()
robe_mtl = wizard2_model.materials['WizardCostume_robe']
robe_mtl['map_Kd'] = 'robe2.png'
wizard2_model.materials['Crystal']['Kd'] = (0.8, 0.0, 0.3)
wizard2_model.materials['Skin_WizardFace2']['map_Kd'] = 'wizard-face2.png'
forest_model = loader.load_model('forest')


MAX_LEVEL = 4
SPELL_LEVELS = {
    1: ['jump', MagicMissile],
    2: ['jump', MagicMissile, Shield],
    3: ['jump', MagicMissile, Fire, Shield],
    4: ['jump', MagicMissile, Fire, Shield, Laser, Invisibility],
}

# Table of which spells can be used against each other
SPELL_REPLIES = {
    'jump': ['jump', MagicMissile, Fire, Shield, Laser],
    'MagicMissile': ['jump', Shield],
    'Shield': [None, Laser],
    'Fire': ['jump', Shield, Fire],
    'Laser': ['jump', Invisibility],
    'Invisibility': [MagicMissile, Fire, Shield, Laser],
    'default': [None] * 4 + [MagicMissile, Fire, Shield],
}


class Body(object):
    def __init__(self, pos, mass):
        self.pos = pos
        self.mass = mass
        self.v = v3((0, 0, 0))
        self.reset_forces()

    def reset_forces(self):
        self.f = v3((0, -GRAVITY * self.mass, 0))
        # self.f = v3((0, 0, 0))

    def apply_force(self, f):
        self.f += f

    def apply_impulse(self, impulse):
        self.v += impulse

    def update(self, dt):
        u = self.v
        self.v += self.f * dt / self.mass
        self.pos += 0.5 * (u + self.v) * dt
        # if the body is on the ground, don't go down further
        if self.pos.y < 0:
            self.pos.y = 0
            self.v.y = 0
        # print dt, self.pos


def action(func):
    """Decorator, ensures that a corpse can't perform actions."""
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if self.dead:
            return
        return func(self, *args, **kwargs)
    return wrapper


class Wizard(pyglet.event.EventDispatcher):
    MASS = 0.001
    dead = False
    jumping = False
    MODEL = wizard_model
    drawn = False

    health = 20

    def __init__(self, pos, rotation):
        self.pos = pos
        self.rotation = rotation
        self.node = ModelNode(self.MODEL, pos, rotation)
        self.body = Body(pos, self.MASS)
        self.spells = []  # maintain a list of the spells cast

    def spawn(self, world):
        self.world = world
        world.objects.append(self)
        world.scene.add(self.node)

    def is_hit(self, p):
        if len(p) == 2:
            x, y = p
            rx, ry, rz = self.receiver.pos
            return (
                rx - 2 < x < rx + 2 and
                ry < y < ry + 6
            )
        else:
            x, y, z = self.pos
            px, py, pz = p
            if y < py < (y + 6):
                dx = x - px
                dz = z - pz
                if dx * dx + dz * dz < 4:
                    return True
            return False

    def get_shield(self):
        for s in self.spells:
            if isinstance(s, Shield):
                return s

    @action
    def hit(self, damage):
        self.health -= damage
        if self.health < 0:
            self.dead = True
            self.node.model_instance.play('death')
            self.dispatch_event('on_death', self)
        else:
            self.node.model_instance.play('hit')

    def kill(self):
        self.world.objects.remove(self)
        self.world.scene.remove(self.node)

    @action
    def draw(self):
        self.node.model_instance.play('draw')
        self.drawn = True

    @action
    def cast(self, spell_cls):
        if not self.drawn:
            self.draw()
        else:
            #print 'casting %s!' % name
            f = spell_cls(self, self.opponent)
            if not f.can_cast():
                return
            f.spawn(self.world)
            self.spells.append(f)
            self.node.model_instance.play('cast')

    def update(self, dt):
        self.body.update(dt)
        #  if you're on the ground, you're not jumping
        if self.body.pos.y == 0:
            self.jumping = False

    @action
    def jump(self):
        #print 'jumping'
        #  you can only jump once
        if not self.jumping:
            self.body.apply_impulse(v3(0, 20, 0))
            self.jumping = True

Wizard.register_event_type('on_death')


class PlayerWizard(Wizard):
    pass


class OpposingWizard(Wizard):
    MODEL = wizard2_model

    def spawn(self, world):
        super(OpposingWizard, self).spawn(world)
        self.ai = AI(self)

    def update_ai(self, dt):
        self.ai.update(dt)


class AI(object):
    def __init__(self, wiz):
        self.wiz = wiz
        self.world = wiz.world
        self.strategy = None

    def pick_strategy(self):
        """Choose a strategy"""
        # if the opponent has drawn, we shoud draw too
        if self.wiz.opponent.drawn:
            if not self.wiz.drawn:
                self.wiz.draw()
                return

        # get opposing spell
        rx_spells = [o for o in self.world.objects
                if isinstance(o, Spell) and
                o.receiver == self.wiz]
        if self.wiz.opponent.jumping:
            rx_spells += ['jump']

        # get list of spells available for the current level
        available_spells = SPELL_LEVELS[self.world.level]

        if len(rx_spells) > 0:
            # print rx_spells, available_spells
            # randomly choose the spell to respond to
            rx_spell = random.choice(rx_spells)
            if rx_spell == 'jump':
                responses = SPELL_REPLIES[rx_spell]
            else:
                responses = SPELL_REPLIES[rx_spell.__class__.__name__]

            available_responses = set(responses).intersection(
                                    set(available_spells)
                                )
            # print 'available_responses:', available_responses
            try:
                self.strategy = random.choice(list(available_responses))
            except IndexError:
                self.strategy = None
        else:
            self.strategy = random.choice(SPELL_REPLIES['default'])

        if self.strategy:
            self.execute_strategy()

    def execute_strategy(self):
        if self.strategy == 'jump':
            self.wiz.jump()
        else:
            self.wiz.cast(self.strategy)

    def update(self, dt):
        self.pick_strategy()


from keybindings import KEYS, JUMP_KEY


class World(pyglet.event.EventDispatcher):
    """collection of all geometry and objects in this world"""
    def __init__(self, scene):
        self.objects = []
        self.scene = scene
        self.level = 1
        self.spellbindings = {}

    def load(self):
        self.scene.clear()
        self.scene.add(Light(
            pos=(0, 1000, -1000, 0),
            colour=(1.0, 1.0, 1.0, 1.0),
            intensity=0.4,
            falloff=0
        ))

        self.wiz = PlayerWizard(
            pos=v3((-15, 0, 0)),
            rotation=(90, 0, 1, 0),
        )
        self.wiz.spawn(self)

        self.wiz2 = OpposingWizard(
            pos=v3((15, 0, 0)),
            rotation=(-90, 0, 1, 0),
        )
        self.wiz2.spawn(self)

        self.wiz.opponent = self.wiz2
        self.wiz2.opponent = self.wiz

        self.forest = ModelNode(forest_model, rotation=(180, 0, 1, 0))
        self.scene.add(self.forest)

        self.particles = ParticleSystemNode()
        self.scene.add(self.particles)
        self.attacks = []

        self.won = False
        self.wiz.set_handler('on_death', self.on_player_death)
        self.wiz2.set_handler('on_death', self.won_level)

    def set_spells(self, spells):
        assert spells
        self.spellbindings = dict(zip(KEYS, spells))

    def set_level(self, level):
        max_level = max(SPELL_LEVELS.keys())
        self.level = min(level, max_level)

    def process_input(self, symbol, modifiers):
        if symbol == JUMP_KEY:
            self.wiz.jump()
        else:
            try:
                spell = self.spellbindings[symbol]
            except KeyError:
                pass
            else:
                self.wiz.cast(spell)

    def update(self, dt):
        for o in self.objects:
            o.update(dt)

    def update_ai(self, dt):
        for o in self.objects:
            if hasattr(o, 'update_ai'):
                o.update_ai(dt)

    def on_player_death(self, wiz):
        self.dispatch_event('game_over')

    def won_level(self, wiz):
        if not self.won:
            # self.show_message('well done!')
            self.won = True
            self.dispatch_event('level_complete')

    def next_level(self):
        if self.level < max(SPELL_LEVELS.keys()):
            self.level += 1
        self.restart()

    def restart(self):
        # clear all spells and wizards
        del_objs = [o for o in self.objects
                    if isinstance(o, Wizard) or isinstance(o, Spell)]
        for o in del_objs:
            o.kill()

        self.scene.objects = []

        # restart the game
        self.load()


World.register_event_type('game_over')
World.register_event_type('level_complete')
