from __future__ import division

import math
import pyglet
from pyglet import gl
import pyglet.resource
from pyglet.window import key
import pyglet.sprite

from scenegraph.scenegraph import Camera, Scene, v3, RenderPass, GLStateGroup, ModelNode
from scenegraph.objloader import ModelLoader
from scenegraph.lighting import DepthOnlyPass, LightingPass, Light, CompositePass
from scenegraph.particles import ParticleDisplayGroup

# Set up pyglet resource loader
pyglet.resource.path += [
    'assets/textures/',
    'assets/sprites/',
    'assets/music/',
    'assets/sounds/'
]
pyglet.resource.reindex()

from duel import World, MAX_LEVEL


WIDTH = 800
HEIGHT = 600
FPS = 60


loader = ModelLoader()


# Pixels the mouse has to move to move the camera 1 degree
MOUSE_SPEED = 4.0

# Theta is left-right rotation around the scene
MAX_THETA = 80
MIN_THETA = -80

# Phi is how high we are looking down on the scene
MIN_PHI = 2
MAX_PHI = 30


level_up_sound = pyglet.resource.media('blessing.ogg', streaming=False)
music = pyglet.resource.media('amure-fantasy.ogg', streaming=True)


class GameState(pyglet.event.EventDispatcher):
    def __init__(self, window, scene):
        self.window = window
        self.scene = scene
        self.zpos = 6
        self.theta = 0
        self.phi = 10
        self.camera = Camera(
            pos=v3((0, 0, 0)),
            look_at=v3((0, 4, 0)),
            width=self.window.width,
            height=self.window.height
        )
        self.update_camera()

        self.world = World(self.scene)
        self.world.load()
        self.world.set_handler('game_over', self.on_game_over)
        self.world.set_handler('level_complete', self.on_level_complete)
        self.message = None

    def on_game_over(self):
        self.show_message('game-over.png')
        pyglet.clock.schedule_once(self.level_end_timer, 4, 'game_over')

    def on_level_complete(self):
        level_up_sound.play()
        if self.world.level == MAX_LEVEL:
            self.show_message('you-win.png')
        else:
            self.show_message('level-up.png')
        pyglet.clock.schedule_once(self.level_end_timer, 4, 'level_complete')

    def level_end_timer(self, dt, event):
        self.dispatch_event(event)

    def on_key_press(self, symbol, modifiers):
        self.world.process_input(symbol, modifiers)

    def on_mouse_motion(self, x, y, dx, dy):
        self.theta = max(MIN_THETA, min(MAX_THETA, self.theta + dx / MOUSE_SPEED))
        self.phi = max(MIN_PHI, min(MAX_PHI, self.phi + dy / MOUSE_SPEED))

    def update_camera(self):
        sinphi = math.sin(math.radians(self.phi))
        cosphi = math.cos(math.radians(self.phi))
        sintheta = math.sin(math.radians(self.theta))
        costheta = math.cos(math.radians(self.theta))
        self.camera.pos = v3(cosphi * sintheta, sinphi, -cosphi * costheta) * (self.zpos + 24)

    def show_message(self, message):
        img = pyglet.resource.image(message)
        self.message = pyglet.sprite.Sprite(img)
        self.message.x = (self.window.width - img.width) // 2
        self.message.y = (self.window.height - img.height) // 2

    def draw_message(self):
        gl.glClear(gl.GL_DEPTH_BUFFER_BIT)

        gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA)
        gl.glDisable(gl.GL_ALPHA_TEST)
        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glLoadIdentity()
        gl.glOrtho(
            0, self.window.width,
            0, self.window.height,
            1, -1
        )
        gl.glMatrixMode(gl.GL_MODELVIEW)
        gl.glLoadIdentity()
        self.message.draw()

    def on_draw(self):
        gl.glClearColor(0, 0, 0, 0)
        gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)

        gl.glDisable(gl.GL_CULL_FACE)

        gl.glEnable(gl.GL_BLEND)
        gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA)
        gl.glEnable(gl.GL_ALPHA_TEST)
        gl.glAlphaFunc(gl.GL_GREATER, 0.9)

        self.scene.render(self.camera)
        if self.message:
            self.draw_message()

    def update(self, dt):
        global zpos
        self.zpos *= 0.8 ** dt
        self.update_camera()
        self.scene.update(dt)
        self.world.update(dt)

    def update_ai(self, dt):
        self.world.update_ai(dt)

    def set_spells(self, spells):
        self.world.set_spells(spells)

    def set_level(self, level):
        self.world.set_level(level)

    def start(self):
        pyglet.clock.schedule_interval(self.update, 1.0 / FPS)
        pyglet.clock.schedule_interval(self.update_ai, 0.5)

        self.window.push_handlers(
            on_draw=self.on_draw,
            on_key_press=self.on_key_press,
            on_mouse_motion=self.on_mouse_motion
        )

    def stop(self):
        pyglet.clock.unschedule(self.update)
        pyglet.clock.unschedule(self.update_ai)
        self.window.pop_handlers()
        self.world.restart()
        self.message = None

GameState.register_event_type('game_over')
GameState.register_event_type('level_complete')


class TitleScreen(pyglet.event.EventDispatcher):
    def __init__(self, window, scene):
        self.window = window
        self.scene = scene
        self.setup_scene()

        self.batch = pyglet.graphics.Batch()
        self.logo_sprite = pyglet.sprite.Sprite(
            pyglet.resource.image('logo.png'),
            batch=self.batch
        )
        self.logo_sprite.position = (20, 90)
        self.start_sprite = pyglet.sprite.Sprite(
            pyglet.resource.image('start.png'),
            batch=self.batch
        )
        from spellselect import Button
        self.button = Button(
            self.start_sprite
        )
        self.start_sprite.position = (175, 40)

    def setup_scene(self):
        from duel import wizard_model, forest_model
        self.scene.clear()
        self.scene.add(Light(
            pos=(1000, 1000, 1000, 0),
            colour=(1.0, 1.0, 1.0, 1.0),
            intensity=0.4,
            falloff=0
        ))
        self.scene.add(Light(
            pos=(10, 0, 10, 0),
            colour=(1.0, 0.2, 0.6, 1.0),
            intensity=300,
            falloff=2
        ))
        self.wizard = ModelNode(wizard_model)
        self.forest = ModelNode(forest_model)
        self.scene.add(self.wizard)
        self.scene.add(self.forest)
        self.camera = Camera(
            pos=v3((-2, 4, 8)),
            look_at=v3((-2, 4, 0)),
            width=self.window.width,
            height=self.window.height
        )

    def start(self):
        self.window.push_handlers(
            on_draw=self.on_draw,
            on_key_press=self.on_key_press,
            on_mouse_press=self.on_mouse_press
        )

    def on_draw(self):
        gl.glClearColor(0, 0, 0, 0)
        gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)

        gl.glDisable(gl.GL_CULL_FACE)

        gl.glEnable(gl.GL_BLEND)
        gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA)
        gl.glEnable(gl.GL_ALPHA_TEST)
        gl.glAlphaFunc(gl.GL_GREATER, 0.9)

        self.scene.render(self.camera)
        gl.glClear(gl.GL_DEPTH_BUFFER_BIT)

        gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA)
        gl.glDisable(gl.GL_ALPHA_TEST)
        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glLoadIdentity()
        gl.glOrtho(
            0, self.window.width,
            0, self.window.height,
            1, -1
        )
        gl.glMatrixMode(gl.GL_MODELVIEW)
        gl.glLoadIdentity()
        self.batch.draw()

    def on_key_press(self, symbol, modifiers):
        if symbol in [key.SPACE, key.ENTER]:
            self.dispatch_event('game_start')

    def on_mouse_press(self, x, y, button, modifiers):
        if self.button.is_mouse_over(x, y):
            self.dispatch_event('game_start')

    def stop(self):
        self.window.pop_handlers()

TitleScreen.register_event_type('game_start')


class Game(object):
    """Control the game. Sets up the world,
    delegates to different game states"""
    def __init__(self, level=1):
        self.window = pyglet.window.Window(width=WIDTH, height=HEIGHT)
        self.level = level
        self.gamestate = None
        self.spells_selector = None
        if level > 1:
            self.show_spell_selector()
        else:
            self.show_title_screen()
        player = music.play()
        player.volume *= 0.5
        player.eos_action = 'loop'

    def get_scene(self):
        """There is only one scene object - we reuse it.

        This helps avoid potentially leaking stuff that isn't properly deallocated.

        """
        try:
            return self.scene
        except AttributeError:
            lighting = LightingPass(
                ambient=(0.05, 0.05, 0.05, 1.0)
            )
            self.scene = Scene([
                lighting,
                CompositePass(lighting),
                RenderPass(
                    transparency=True,
                    group=ParticleDisplayGroup(),
                )
            ])
            return self.scene

    def show_title_screen(self):
        self.title_screen = TitleScreen(self.window, self.get_scene())
        self.title_screen.set_handler('game_start', self.on_game_start)
        self.title_screen.start()

    def on_game_start(self):
        self.title_screen.stop()
        self.show_spell_selector()

    def show_spell_selector(self):
        from spellselect import SpellSelector
        self.spells_selector = SpellSelector(self.window, level=self.level)
        self.spells_selector.set_handler('spells_chosen', self.on_spells_chosen)
        self.spells_selector.start()

    def on_spells_chosen(self, spells):
        self.spells_selector.stop()
        self.spells_selector = None
        self.start_game(spells)

    def start_game(self, spells):
        if self.gamestate is None:
            self.gamestate = GameState(self.window, self.get_scene())
            self.gamestate.set_handler('game_over', self.game_over)
            self.gamestate.set_handler('level_complete', self.level_complete)
        else:
            self.gamestate.world.restart()

        self.gamestate.set_spells(spells)
        self.gamestate.set_level(self.level)
        self.gamestate.start()

    def game_over(self):
        self.level = 1
        self.gamestate.stop()
        self.show_title_screen()

    def level_complete(self):
        self.gamestate.stop()

        self.level += 1
        if self.level > MAX_LEVEL:
            self.level = 1
            self.show_title_screen()
        else:
            self.show_spell_selector()
