import lepton
from lepton.emitter import StaticEmitter, PerParticleEmitter
from lepton import Particle
from lepton import domain
from lepton import controller
from lepton.domain import Vec3

from pyglet import gl
import pyglet.resource
from euclid import Vector3
from scenegraph.objloader import load_texture, Model
from scenegraph.lighting import Light
from scenegraph.scenegraph import ModelNode, ShaderGroup, GLStateGroup, RayNode
from scenegraph.sphere import Sphere
from scenegraph.shader import Shader


fire_sound = pyglet.resource.media('flamethrower.ogg', streaming=False)
laser_sound = pyglet.resource.media('teleport.ogg', streaming=False)
shield_on_sound = pyglet.resource.media('moving.ogg', streaming=False)
shield_off_sound = pyglet.resource.media('spring.ogg', streaming=False)
mm_sound = pyglet.resource.media('warp.ogg', streaming=False)
invisibility_sound = pyglet.resource.media('enchant.ogg', streaming=False)

fireball = load_texture('fire.png')
magic_missile = load_texture('magic-missile.png')
spark = load_texture('spark.png')


class Spell(object):
    """Base class for spells, sets up the firing direction."""
    MAX_PER_CASTER = 1

    def __init__(self, caster, receiver):
        self.caster = caster
        self.receiver = receiver
        x, y, z = (receiver.pos - caster.pos)
        self.direction = Vector3(x, 0, z).normalized()

    def can_cast(self):
        num = 0
        for s in self.caster.spells:
            if isinstance(s, Shield):
                return False
            if isinstance(s, self.__class__):
                num += 1
        return num < self.MAX_PER_CASTER

    def handle_shield(self):
        shield = self.receiver.get_shield()
        if shield:
            self.set_receiver_shielded(shield)
        else:
            self.set_receiver_unshielded()

    def set_receiver_shielded(self, shield):
        pass

    def set_receiver_unshielded(self):
        pass

    def get_origin_point(self):
        """Get the start point for spell effects."""
        return self.caster.pos + Vector3(0, 4, 0) + self.direction * 2

    def get_target_point(self):
        """Get the point to aim for."""
        x, y, z = self.receiver.pos
        return Vector3(x, self.caster.pos.y + 4, z)

    def kill(self):
        try:
            self.caster.spells.remove(self)
        except ValueError:
            pass
        try:
            self.world.objects.remove(self)
        except ValueError:
            pass


class Fire(Spell):
    MAX_PER_CASTER = 2
    HIT_PARTICLES = 6  # number of particles used to test hit
    DAMAGE = 2  # per hit particle

    def spawn(self, world):
        self.world = world
        self.scene = world.scene
        world.objects.append(self)

        self.particle_group = world.particles.create_group(texture=fireball, controllers=[
            controller.Movement(),
            controller.Fader(
                start_alpha=0.0,
                fade_in_start=0.0,
                fade_in_end=0.2,
                max_alpha=1.0,
                fade_out_start=1.0,
                fade_out_end=1.5,
                end_alpha=0.0
            ),
            controller.Lifetime(1.5),
            controller.Growth(1.5)
        ])

        pos = self.get_origin_point()
        pos_domain = domain.Cone(
            pos,
            pos + self.direction * 3,
            0.5
        )

        template = Particle(
            color=(1, 1, 1),
            velocity=tuple(self.direction * 25)
        )

        emitter = StaticEmitter(
            position=pos_domain,
            velocity=domain.Sphere(tuple(self.direction * 25), 4),
            size=domain.Line((0.1, 0.1, 0.1), (1.0, 1.0, 1.0)),
            template=template,
            rate=200,
            time_to_live=0.5,
        )

        self.particle_group.bind_controller(emitter)

        self.light = Light(
            pos=pos,
            colour=(1.0, 0.3, 0.1, 1.0),
            intensity=0,
            falloff=2
        )

        self.age = 0
        self.scene.add(self.light)

        self.hit_particles = [
            pos + self.direction * i for i in range(self.HIT_PARTICLES)
        ]
        self.handle_shield()
        fire_sound.play()

    def set_receiver_shielded(self, shield):
        self.shield_bounce = controller.Bounce(
            domain.Sphere(tuple(shield.pos), shield.RADIUS),
            bounce=0
        )
        self.receiver_shield = shield
        self.particle_group.bind_controller(self.shield_bounce)

    def set_receiver_unshielded(self):
        try:
            self.particle_group.unbind_controller(self.shield_bounce)
        except AttributeError:
            pass
        self.receiver_shield = None

    def kill(self):
        Spell.kill(self)
        self.scene.remove(self.light)
        self.particle_group.system.remove_group(self.particle_group)

    def test_hit(self, dt):
        is_hit = self.receiver.is_hit
        surviving = []
        shield = self.receiver_shield

        for p in self.hit_particles:
            p += self.direction * 25 * dt
            if shield and shield.is_hit(p):
                continue

            if is_hit(p):
                self.receiver.hit(self.DAMAGE)
            else:
                surviving.append(p)
        self.hit_particles = surviving

    def update(self, dt):
        self.test_hit(dt)

        self.age += dt
        self.light.pos += self.direction * 15 * dt
        if self.age < 1.0:
            self.light.intensity += 3000 * dt
        elif self.age < 2.5:
            self.light.intensity *= 0.2 ** dt
        else:
            self.kill()


class MagicMissile(Spell):
    DAMAGE = 5
    MAX_PER_CASTER = 2

    def spawn(self, world):
        self.world = world
        self.scene = world.scene
        world.objects.append(self)

        pos = self.get_origin_point()
        target = self.get_target_point()

        self.missile_group = world.particles.create_group(
            texture=magic_missile,
            controllers=[
                controller.Movement(),
                controller.Lifetime(1.0),
                controller.Magnet(
                    domain.Sphere(target, 1),
                    100,
                    0
                ),
                controller.Fader(
                    start_alpha=0.0,
                    fade_in_start=0.0,
                    fade_in_end=0.2,
                    max_alpha=1.0,
                ),
                controller.Growth(5, 0.96)
        ]
        )

        pos_domain = domain.Cone(
            pos,
            pos + tuple(self.direction * 3),
            0.5
        )

        emitter = StaticEmitter(
            template=Particle(
                rotation=(0, 0, 6),
                color=(1, 1, 1, 1),
                velocity=tuple(self.direction * 15),
                size=(0.1, 0.1, 0.1),
            ),
            position=pos_domain,
            velocity=domain.Disc(tuple(self.direction * 10), (1, 0, 0), 7, 5),
        )

        emitter.emit(1, self.missile_group)

        self.spark_group = world.particles.create_group(
            texture=spark,
            controllers=[
                controller.Movement(),
                controller.Gravity((0, 1, 0)),
                controller.Lifetime(0.8),
                controller.Growth(0.1),
                controller.Fader(
                    start_alpha=0.0,
                    fade_in_start=0.0,
                    fade_in_end=0.035,
                    max_alpha=1.0,
                    fade_out_start=0.2,
                    fade_out_end=0.8,
                    end_alpha=0.0
                ),
            ]
        )

        per_particle_emitter = PerParticleEmitter(
            self.missile_group,
            template=Particle(
                size=(0.2, 0.2, 0.2),
                color=(0.6, 1.0, 0.8, 1.0),
            ),
            deviation=Particle(
                position=(0.1, 0.1, 0.1),
    #            color=(0.1, 0.1, 0.1, 0),
            ),
            rate=50,
            rotation=domain.Line((0, 0, 0), (0, 0, 6)),
            up=domain.Line((0, 0, 0), (0, 0, 1))
        )

        self.spark_group.bind_controller(per_particle_emitter)

        self.light = Light(
            pos=pos,
            colour=(0.1, 1.0, 0.3, 1.0),
            intensity=8000,
            falloff=4
        )

        self.age = 0
        self.v = 15
        self.scene.add(self.light)
        self.handle_shield()
        mm_sound.play()

    def set_receiver_shielded(self, shield):
        self.shield_bounce = controller.Bounce(
            domain.Sphere(tuple(shield.pos), shield.RADIUS),
            bounce=0.7
        )
        self.missile_group.bind_controller(self.shield_bounce)

    def set_receiver_unshielded(self):
        try:
            self.missile_group.unbind_controller(self.shield_bounce)
        except AttributeError:
            pass

    def kill(self):
        Spell.kill(self)
        self.scene.remove(self.light)
        self.missile_group.system.remove_group(self.missile_group)
        self.spark_group.system.remove_group(self.spark_group)

    def update(self, dt):
        self.age += dt
        is_hit = self.receiver.is_hit
        for particle in self.missile_group:
            p = particle.position
            if is_hit(p):
                self.missile_group.kill(particle)
                self.receiver.hit(self.DAMAGE)
                self.kill()
                return
            else:
                self.light.pos = tuple(p)
        if self.age > 1.0:
            self.kill()


class Shield(Spell):
    RADIUS = 7

    model = Model(
        meshes=[Sphere(RADIUS, 20, 40)]
    )
    shader = Shader(
        vert="""
varying vec3 normal;
varying vec3 eye;

void main(void)
{
    gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;
    normal = (gl_NormalMatrix * gl_Normal).xyz;
    eye = -vec3(gl_ModelViewMatrix * gl_Vertex);
}
""",
        frag="""
varying vec3 normal;
varying vec3 eye;

uniform vec3 colour;
uniform float power;

void main (void) {
    float frac = pow(dot(normalize(normal), normalize(eye)), power);
    gl_FragColor = vec4(colour, 1.0 - frac);
}
""")

    def spawn(self, world):
        self.world = world
        self.scene = world.scene
        world.objects.append(self)

        pos = self.get_origin_point()
        self.light = Light(
            pos=pos,
            colour=(0.3, 0.3, 1.0, 1.0),
            intensity=4000,
            falloff=2
        )

        self.age = 0
        self.scene.add(self.light)

        glgroup = GLStateGroup(
            disable=[gl.GL_ALPHA_TEST],
            cull_face=gl.GL_FRONT
        )
        self.group = ShaderGroup(self.shader, parent=glgroup)
        self.group.uniformf('colour', 0.3, 0.8, 1.0)
        self.group.uniformf('power', 2)
        self.node = ModelNode(self.model, pos=pos, group=self.group, transparent=True)
        self.scene.add(self.node)
        self.pos = pos
        self.fire_receiver_shielded()
        shield_on_sound.play()

    def fire_receiver_shielded(self):
        """Call the set_receiver_shielded callback on opposing spells."""
        for s in self.receiver.spells:
            s.set_receiver_shielded(self)

    def fire_receiver_unshielded(self):
        """Call the set_receiver_unshielded callback on opposing spells."""
        for s in self.receiver.spells:
            s.set_receiver_unshielded()

    def kill(self):
        Spell.kill(self)
        self.scene.remove(self.light)
        self.scene.remove(self.node)
        self.fire_receiver_unshielded()

    def is_hit(self, p):
        """Test whether a point has hit the shield.

        To do this we basically make the shield a thick shell and test whether
        the point is in it. This is crude but works for particles that aren't
        moving too fast.

        """
        o = self.RADIUS
        o2 = o * o
        i = self.RADIUS - 1
        i2 = i * i
        return i2 < (p - self.pos).magnitude_squared() < o2

    def update(self, dt):
        self.age += dt
        self.group.uniformf('power', 2.5 - self.age)
        if self.age < 2.5:
            self.light.intensity *= 0.4 ** dt
        else:
            shield_off_sound.play()
            self.kill()


class Laser(Spell):
    DAMAGE = 8
    shader = Shader(
        vert="""
varying vec3 normal;
varying vec3 eye;

void main(void)
{
    gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;
    normal = (gl_NormalMatrix * gl_Normal).xyz;
    eye = -vec3(gl_ModelViewMatrix * gl_Vertex);
}
""",
        frag="""
varying vec3 normal;
varying vec3 eye;

uniform vec3 colour;
uniform vec3 colour2;

void main (void) {
    gl_FragColor = vec4(colour, 0.5);
}
""")

    def spawn(self, world):
        self.world = world
        self.scene = world.scene
        world.objects.append(self)

        pos = self.get_origin_point()
        self.light = Light(
            pos=pos,
            colour=(0.8, 0.0, 0.3, 1.0),
            intensity=4000,
            falloff=2
        )
        self.scene.add(self.light)

        self.age = 0

        glgroup = GLStateGroup(
            disable=[gl.GL_ALPHA_TEST],
            cull_face=gl.GL_BACK
        )
        self.group = ShaderGroup(self.shader, parent=glgroup)
        self.group.uniformf('colour', 0.8, 0.0, 0.3)

        p1 = pos
        p2 = pos + self.direction * 50
        self.node = RayNode(
            p1, p2,
            0.1,
            group=self.group, transparent=True
        )
        self.scene.add(self.node)

        self.spark_group = world.particles.create_group(
            texture=spark,
            controllers=[
                controller.Movement(),
                controller.Gravity((0, 1, 0)),
                controller.Lifetime(0.8),
                controller.Growth(0.1),
                controller.Fader(
                    start_alpha=0.0,
                    fade_in_start=0.0,
                    fade_in_end=0.035,
                    max_alpha=1.0,
                    fade_out_start=0.2,
                    fade_out_end=0.8,
                    end_alpha=0.0
                ),
            ]
        )

        self.emitter = None
        self.update_emitter()
        laser_sound.play()

    def update_emitter(self):
        if self.emitter:
            self.spark_group.unbind_controller(self.emitter)

        self.emitter = StaticEmitter(
            position=domain.Line(tuple(self.node.p1), tuple(self.node.p2)),
            template=Particle(
                size=(0.3, 0.3, 0.3),
                color=(0.8, 0, 0.4, 1.0),
                velocity=(0, 1, 0),
            ),
            rate=50,
            rotation=domain.Line((0, 0, 0), (0, 0, 6)),
            up=domain.Line((0, 0, 0), (0, 0, 1))
        )
        self.spark_group.bind_controller(self.emitter)

    def kill(self):
        Spell.kill(self)
        self.scene.remove(self.light)
        self.scene.remove(self.node)
        self.spark_group.system.remove_group(self.spark_group)

    def update(self, dt):
        pos = self.get_origin_point()
        self.node.p1 = pos
        if (self.receiver.dead or pos.y < self.receiver.pos.y or
                pos.y > self.receiver.pos.y + 6 or
                any(isinstance(s, Invisibility) for s in self.receiver.spells)):
            self.node.p2 = pos + self.direction * 50
        else:
            self.node.p2 = self.get_target_point()
            self.receiver.hit(self.DAMAGE * dt)
        self.update_emitter()

        self.light.pos = pos

        self.age += dt
        if self.age < 0.4:
            self.node.width += 0.5 * dt
        elif self.age > 0.8:
            self.node.width *= 0.4 ** dt
            self.light.intensity *= 0.4 ** dt
        if self.age > 1.5:
            self.kill()


class Invisibility(Spell):
    shader = Shader(
        vert="""
varying vec3 normal;
varying vec3 eye;
varying vec2 uv;

void main(void)
{
    gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;
    normal = (gl_NormalMatrix * gl_Normal).xyz;
    eye = -vec3(gl_ModelViewMatrix * gl_Vertex);
    uv = gl_MultiTexCoord0.st;
}
""",
        frag="""
varying vec3 normal;
varying vec3 eye;
varying vec2 uv;

uniform float power;
uniform sampler2D diffuse;

void main (void) {
    float frac = pow(dot(normalize(normal), normalize(eye)), power);
    vec4 mapcolour = texture2D(diffuse, uv);
    gl_FragColor = vec4(mapcolour.rgb, 1.0 - frac);
}
""")
    shader.bind_material_to_uniformf('Kd', 'colour')
    shader.bind_material_to_texture('map_Kd', 'diffuse')

    def spawn(self, world):
        self.world = world
        self.scene = world.scene
        world.objects.append(self)
        self.caster.node.transparent = True
        self.orig_group = self.caster.node.group
        self.group = ShaderGroup(self.shader)
        self.caster.node.group = self.group
        self.group.uniformf('power', 0.25)
        self.age = 0
        invisibility_sound.play()

    def kill(self):
        Spell.kill(self)
        self.caster.node.transparent = False
        self.caster.node.group = self.orig_group

    def update(self, dt):
        self.age += dt
        if self.age > 2.5:
            invisibility_sound.play()
            self.kill()
