from functools import partial
import pyglet
import pyglet.graphics
import pyglet.sprite
from pyglet import gl
import pyglet.event

from game import WIDTH, HEIGHT
import spells


OFFENSIVE_SPELLS = [
    {
        'level': 1,
        'name': 'Magic Missile',
        'icon': 'magic-missile-icon.png',
        'description': 'A simple charm. Evoke energy and fling it at your nemesis(es) in the shape of a ball.',
        'anchor': (8, 8),
        'class': spells.MagicMissile,
    },
    {
        'level': 2,
        'name': 'Throw Flame',
        'icon': 'fire-icon.png',
        'description': 'Channel volcano heat from the centre of the earth to attack your foes.',
        'anchor': (16, 7),
        'class': spells.Fire,
    },
    {
        'level': 3,
        'name': 'Light Beam',
        'icon': 'laser.png',
        'description': 'Amplify light energy into a powerful beam through stimulated emission of electromagnetic radiation.',
        'anchor': (9, 8),
        'class': spells.Laser,
    }
]

DEFENSIVE_SPELLS = [
    {
        'level': 1,
        'name': 'Mage Jump',
        'icon': 'jump.png',
        'description': "When you want to get out of harm's way fast, there's no place like the place just above your head.\n\n(You always have this).",
        'class': None,
    },
    {
        'level': 2,
        'name': 'Arcane Shield',
        'icon': 'shield.png',
        'description': "Through deep concentration the conjuror can create a barrier of pure arcane energy.",
        'class': spells.Shield,
    },
    {
        'level': 3,
        'name': 'Invisibility',
        'icon': 'invisibility.png',
        'description': "Warlocks who use this spell can move partly into another dimension, reducing their presence in our realm.",
        'class': spells.Invisibility,
    },
]


class Button(object):
    def __init__(self,
            sprite,
            on_mouse_over=None,
            on_click=None,
            selected=False,
            disabled=False):

        self.sprite = sprite
        self.selected = selected
        self.disabled = disabled
        self.on_mouse_over = on_mouse_over
        self.on_click = on_click
        self.set_highlight(False)

    def is_mouse_over(self, x, y):
        sx, sy = self.sprite.position
        return (
            sx < x < sx + self.sprite.width and
            sy < y < sy + self.sprite.height
        )

    def set_highlight(self, highlight):
        if self.disabled:
            self.sprite.opacity = 64
        elif highlight or self.selected:
            self.sprite.opacity = 255
        else:
            self.sprite.opacity = 128
        self.highlight = highlight

    def fire_click(self):
        if self.on_click and not self.disabled:
            self.on_click()

    def set_selected(self, selected):
        self.selected = selected
        self.set_highlight(self.highlight)

    def set_disabled(self, disabled):
        self.disabled = disabled
        self.set_highlight(self.highlight)


class JumpIcon(Button):
    """jump can't be deselected."""
    def set_highlight(self, highlight):
        pass


class SpellSelector(pyglet.event.EventDispatcher):
    def __init__(self, window=None, level=1):
        self.window = window
        if not self.window:
            self.window = pyglet.window.Window(
                width=WIDTH,
                height=HEIGHT
            )

        self.level = level
        self.slots = level
        self.batch = pyglet.graphics.Batch()

        self.logo = self.get_icon('spellbook.png')
        self.logo.position = 20, 20

        self.spell_name = pyglet.text.Label(
            '',
            font_name='Tahoma',
            font_size=26,
            x=self.window.width // 2,
            y=self.window.height - 80,
            color=(200, 150, 40, 255),
            anchor_x='center',
            batch=self.batch
        )
        self.level_name = pyglet.text.Label(
            'Level %d' % self.level,
            font_name='Tahoma',
            font_size=26,
            x=275,
            y=33,
            color=(200, 150, 40, 255),
            batch=self.batch
        )
        self.spell_info = pyglet.text.Label(
            '',
            font_name='Tahoma',
            font_size=18,
            x=self.window.width // 2,
            y=self.window.height - 150,
            width=400,
            color=(200, 200, 200, 255),
            anchor_x='center',
            multiline=True,
            batch=self.batch
        )

        self.selected = set()

        done = self.get_icon('done-button.png')
        done.position = self.window.width - done.width - 20, 30
        self.done_button = Button(
            done,
            on_click=self.on_done,
            disabled=True
        )

        self.layout()

    def get_icon(self, name, anchor=None):
        img = pyglet.resource.image(name)
        if anchor:
            img.anchor_x, img.anchor_y = anchor
        return pyglet.sprite.Sprite(
            img,
            batch=self.batch
        )

    def get_buttons(self):
        return self.spells + [self.done_button]

    def on_mouse_motion(self, x, y, dx, dy):
        current = None
        for s in self.get_buttons():
            if s.is_mouse_over(x, y):
                s.set_highlight(True)
                current = s
            else:
                s.set_highlight(False)

        if current and current.on_mouse_over:
            current.on_mouse_over()

    def on_mouse_press(self, x, y, button, modifiers):
        for s in self.get_buttons():
            if s.is_mouse_over(x, y):
                s.fire_click()
                break

    def set_display_spell(self, spell):
        self.spell_name.text = spell['name']
        self.spell_info.text = spell['description']

    def set_ready(self, ready):
        if ready:
            self.disable_spells()
            self.done_button.set_disabled(False)
        else:
            self.enable_spells()
            self.done_button.set_disabled(True)

    def disable_spells(self):
        for s in self.spells:
            s.set_disabled(not s.selected)

    def enable_spells(self):
        for s in self.spells:
            s.set_disabled(False)

    def toggle_spell(self, spell):
        cls = spell['class']
        button = spell['button']
        if cls in self.selected:
            self.selected.remove(cls)
            button.set_selected(False)
            if len(self.selected) < self.slots:
                self.set_ready(False)
        elif len(self.selected) < self.slots:
            self.selected.add(cls)
            button.set_selected(True)
            if len(self.selected) == self.slots:
                self.set_ready(True)

    def get_spell_button(self, spell, x, y):
        s = self.get_icon(spell['icon'], anchor=spell.get('anchor'))
        s.position = x, y
        if spell['class'] is not None:
            b = Button(s,
                on_mouse_over=partial(self.set_display_spell, spell),
                on_click=partial(self.toggle_spell, spell),
                selected=spell['class'] in self.selected
            )
            spell['button'] = b
        else:
            b = Button(s,
                on_mouse_over=partial(self.set_display_spell, spell),
                selected=True
            )
        return b

    def set_level(self, level):
        self.level = level
        self.slots = min(level, 4)
        if self.slots < len(self.selected):
            self.selected = set()
        self.level_name.text = 'Level %d' % self.level
        self.layout()

    def layout(self):
        self.spells = []
        for i, spell in enumerate(OFFENSIVE_SPELLS):
            if spell['level'] > self.level:
                continue
            self.spells.append(
                self.get_spell_button(spell, 50, 200 + 100 * i)
            )

        for i, spell in enumerate(DEFENSIVE_SPELLS):
            if spell['level'] > self.level:
                continue
            self.spells.append(
                self.get_spell_button(spell, self.window.width - 100, 200 + 100 * i)
            )

    def on_done(self):
        self.dispatch_event('spells_chosen', self.selected)

    def on_draw(self):
        gl.glClearColor(0, 0, 0, 255)
        gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glLoadIdentity()
        gl.glOrtho(
            0, self.window.width,
            0, self.window.height,
            1, -1
        )
        gl.glMatrixMode(gl.GL_MODELVIEW)
        gl.glLoadIdentity()
        self.batch.draw()

    def start(self):
        self.window.push_handlers(
            on_draw=self.on_draw,
            on_mouse_motion=self.on_mouse_motion,
            on_mouse_press=self.on_mouse_press
        )

    def stop(self):
        self.window.pop_handlers()


SpellSelector.register_event_type('spells_chosen')


if __name__ == '__main__':
    import sys
    try:
        level = int(sys.argv[1])
    except:
        level = 3
    s = SpellSelector(level=level)
    s.start()
    pyglet.app.run()
